//
// GLSL textureless classic 4D noise "cnoise",
// with an RSL-style periodic variant "pnoise".
// Author:  Stefan Gustavson (stefan.gustavson@liu.se)
// Version: 2011-08-22
//
// Many thanks to Ian McEwan of Ashima Arts for the
// ideas for permutation and gradient selection.
//
// Copyright (c) 2011 Stefan Gustavson. All rights reserved.
// Distributed under the MIT license. See LICENSE file.
// https://github.com/ashima/webgl-noise
//

import 'package:vector_math/vector_math.dart';

final Vector4 ones = new Vector4.all(1.0);
final Vector4 halves = new Vector4.all(0.5);
final Vector4 zeros = new Vector4.zero();

Vector4 mod289(Vector4 x) {
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

Vector4 mult(Vector4 v0, Vector4 v1) {
  final out = v0.clone();
  out.multiply(v1);
  return out;
}

double dot(Vector4 v0, Vector4 v1) {
  return v0.dot(v1);
}

Vector4 floor(Vector4 v) {
  final out = v.clone();
  out.floor();
  return out;
}

Vector4 abs(Vector4 v) {
  final out = v.clone();
  out.absolute();
  return out;
}

Vector4 step(Vector4 e, Vector4 v) {
  final out = new Vector4.zero();
  for (int i = 0; i < 4; i++) out[i] = (v[i] < e[i]) ? 0.0 : 1.0;
  return out;
}

Vector4 mod(Vector4 v0, Vector4 v1) {
  final out = new Vector4.zero();
  for (int i = 0; i < 4; i++) out[i] = v0[i] % v1[i];
  return out;
}

Vector4 fract(Vector4 v) {
  return v - floor(v);
}

Vector4 permute(Vector4 x) {
  final out = x.clone();
  out.scale(34.0);
  out.add(ones);
  out.multiply(x);
  return mod289(out);
}

Vector4 taylorInvSqrt(Vector4 r) {
  var out = new Vector4.all(1.79284291400159);
  r.scale(0.85373472095314);
  return out - r;
}

Vector4 fade(Vector4 t) {
  var out = t.clone();
  out.scale(6.0);
  out.sub(new Vector4.all(15.0));
  out.multiply(t);
  out.add(new Vector4.all(10.0));
  out.multiply(t);
  out.multiply(t);
  out.multiply(t);
  return out;
}

// Classic Perlin noise
double ashima_cnoise4(Vector4 P) {
  Vector4 Pi0 = floor(P); // Integer part for indexing
  Vector4 Pi1 = Pi0 + ones; // Integer part + 1
  Pi0 = mod289(Pi0);
  Pi1 = mod289(Pi1);
  Vector4 Pf0 = fract(P); // Fractional part for interpolation
  Vector4 Pf1 = Pf0 - ones; // Fractional part - 1.0
  Vector4 ix = new Vector4(Pi0.x, Pi1.x, Pi0.x, Pi1.x);
  Vector4 iy = new Vector4(Pi0.y, Pi0.y, Pi1.y, Pi1.y);
  Vector4 iz0 = Pi0.zzzz;
  Vector4 iz1 = Pi1.zzzz;
  Vector4 iw0 = Pi0.wwww;
  Vector4 iw1 = Pi1.wwww;

  Vector4 ixy = permute(permute(ix) + iy);
  Vector4 ixy0 = permute(ixy + iz0);
  Vector4 ixy1 = permute(ixy + iz1);
  Vector4 ixy00 = permute(ixy0 + iw0);
  Vector4 ixy01 = permute(ixy0 + iw1);
  Vector4 ixy10 = permute(ixy1 + iw0);
  Vector4 ixy11 = permute(ixy1 + iw1);

  Vector4 gx00 = ixy00 * (1.0 / 7.0);
  Vector4 gy00 = floor(gx00) * (1.0 / 7.0);
  Vector4 gz00 = floor(gy00) * (1.0 / 6.0);
  gx00 = fract(gx00) - halves;
  gy00 = fract(gy00) - halves;
  gz00 = fract(gz00) - halves;
  Vector4 gw00 = new Vector4.all(0.75) - abs(gx00) - abs(gy00) - abs(gz00);
  Vector4 sw00 = step(gw00, zeros);
  gx00 -= mult(sw00, (step(zeros, gx00) - halves));
  gy00 -= mult(sw00, (step(ones, gy00) - halves));

  Vector4 gx01 = ixy01 * (1.0 / 7.0);
  Vector4 gy01 = floor(gx01) * (1.0 / 7.0);
  Vector4 gz01 = floor(gy01) * (1.0 / 6.0);
  gx01 = fract(gx01) - halves;
  gy01 = fract(gy01) - halves;
  gz01 = fract(gz01) - halves;
  Vector4 gw01 = new Vector4.all(0.75) - abs(gx01) - abs(gy01) - abs(gz01);
  Vector4 sw01 = step(gw01, zeros);
  gx01 -= mult(sw01, (step(zeros, gx01) - halves));
  gy01 -= mult(sw01, (step(zeros, gy01) - halves));

  Vector4 gx10 = ixy10 * (1.0 / 7.0);
  Vector4 gy10 = floor(gx10) * (1.0 / 7.0);
  Vector4 gz10 = floor(gy10) * (1.0 / 6.0);
  gx10 = fract(gx10) - halves;
  gy10 = fract(gy10) - halves;
  gz10 = fract(gz10) - halves;
  Vector4 gw10 = new Vector4.all(0.75) - abs(gx10) - abs(gy10) - abs(gz10);
  Vector4 sw10 = step(gw10, zeros);
  gx10 -= mult(sw10, (step(zeros, gx10) - halves));
  gy10 -= mult(sw10, (step(zeros, gy10) - halves));

  Vector4 gx11 = ixy11 * (1.0 / 7.0);
  Vector4 gy11 = floor(gx11) * (1.0 / 7.0);
  Vector4 gz11 = floor(gy11) * (1.0 / 6.0);
  gx11 = fract(gx11) - halves;
  gy11 = fract(gy11) - halves;
  gz11 = fract(gz11) - halves;
  Vector4 gw11 = new Vector4.all(0.75) - abs(gx11) - abs(gy11) - abs(gz11);
  Vector4 sw11 = step(gw11, zeros);
  gx11 -= mult(sw11, (step(zeros, gx11) - halves));
  gy11 -= mult(sw11, (step(zeros, gy11) - halves));

  Vector4 g0000 = new Vector4(gx00.x, gy00.x, gz00.x, gw00.x);
  Vector4 g1000 = new Vector4(gx00.y, gy00.y, gz00.y, gw00.y);
  Vector4 g0100 = new Vector4(gx00.z, gy00.z, gz00.z, gw00.z);
  Vector4 g1100 = new Vector4(gx00.w, gy00.w, gz00.w, gw00.w);
  Vector4 g0010 = new Vector4(gx10.x, gy10.x, gz10.x, gw10.x);
  Vector4 g1010 = new Vector4(gx10.y, gy10.y, gz10.y, gw10.y);
  Vector4 g0110 = new Vector4(gx10.z, gy10.z, gz10.z, gw10.z);
  Vector4 g1110 = new Vector4(gx10.w, gy10.w, gz10.w, gw10.w);
  Vector4 g0001 = new Vector4(gx01.x, gy01.x, gz01.x, gw01.x);
  Vector4 g1001 = new Vector4(gx01.y, gy01.y, gz01.y, gw01.y);
  Vector4 g0101 = new Vector4(gx01.z, gy01.z, gz01.z, gw01.z);
  Vector4 g1101 = new Vector4(gx01.w, gy01.w, gz01.w, gw01.w);
  Vector4 g0011 = new Vector4(gx11.x, gy11.x, gz11.x, gw11.x);
  Vector4 g1011 = new Vector4(gx11.y, gy11.y, gz11.y, gw11.y);
  Vector4 g0111 = new Vector4(gx11.z, gy11.z, gz11.z, gw11.z);
  Vector4 g1111 = new Vector4(gx11.w, gy11.w, gz11.w, gw11.w);

  Vector4 norm00 = taylorInvSqrt(new Vector4(dot(g0000, g0000),
      dot(g0100, g0100), dot(g1000, g1000), dot(g1100, g1100)));
  g0000 *= norm00.x;
  g0100 *= norm00.y;
  g1000 *= norm00.z;
  g1100 *= norm00.w;

  Vector4 norm01 = taylorInvSqrt(new Vector4(dot(g0001, g0001),
      dot(g0101, g0101), dot(g1001, g1001), dot(g1101, g1101)));
  g0001 *= norm01.x;
  g0101 *= norm01.y;
  g1001 *= norm01.z;
  g1101 *= norm01.w;

  Vector4 norm10 = taylorInvSqrt(new Vector4(dot(g0010, g0010),
      dot(g0110, g0110), dot(g1010, g1010), dot(g1110, g1110)));
  g0010 *= norm10.x;
  g0110 *= norm10.y;
  g1010 *= norm10.z;
  g1110 *= norm10.w;

  Vector4 norm11 = taylorInvSqrt(new Vector4(dot(g0011, g0011),
      dot(g0111, g0111), dot(g1011, g1011), dot(g1111, g1111)));
  g0011 *= norm11.x;
  g0111 *= norm11.y;
  g1011 *= norm11.z;
  g1111 *= norm11.w;

  double n0000 = dot(g0000, Pf0);
  double n1000 = dot(g1000, new Vector4(Pf1.x, Pf0.y, Pf0.z, Pf0.w));
  double n0100 = dot(g0100, new Vector4(Pf0.x, Pf1.y, Pf0.z, Pf0.w));
  double n1100 = dot(g1100, new Vector4(Pf1.x, Pf1.y, Pf0.z, Pf0.w));
  double n0010 = dot(g0010, new Vector4(Pf0.x, Pf0.y, Pf1.z, Pf0.w));
  double n1010 = dot(g1010, new Vector4(Pf1.x, Pf0.y, Pf1.z, Pf0.w));
  double n0110 = dot(g0110, new Vector4(Pf0.x, Pf1.y, Pf1.z, Pf0.w));
  double n1110 = dot(g1110, new Vector4(Pf1.x, Pf1.y, Pf1.z, Pf0.w));
  double n0001 = dot(g0001, new Vector4(Pf0.x, Pf0.y, Pf0.z, Pf1.w));
  double n1001 = dot(g1001, new Vector4(Pf1.x, Pf0.y, Pf0.z, Pf1.w));
  double n0101 = dot(g0101, new Vector4(Pf0.x, Pf1.y, Pf0.z, Pf1.w));
  double n1101 = dot(g1101, new Vector4(Pf1.x, Pf1.y, Pf0.z, Pf1.w));
  double n0011 = dot(g0011, new Vector4(Pf0.x, Pf0.y, Pf1.z, Pf1.w));
  double n1011 = dot(g1011, new Vector4(Pf1.x, Pf0.y, Pf1.z, Pf1.w));
  double n0111 = dot(g0111, new Vector4(Pf0.x, Pf1.y, Pf1.z, Pf1.w));
  double n1111 = dot(g1111, Pf1);

  Vector4 fade_xyzw = fade(Pf0);
  Vector4 n_0w = new Vector4.zero();
  Vector4.mix(new Vector4(n0000, n1000, n0100, n1100),
      new Vector4(n0001, n1001, n0101, n1101), fade_xyzw.w, n_0w);
  Vector4 n_1w = new Vector4.zero();
  Vector4.mix(new Vector4(n0010, n1010, n0110, n1110),
      new Vector4(n0011, n1011, n0111, n1111), fade_xyzw.w, n_1w);
  Vector4 n_zw = new Vector4.zero();
  Vector4.mix(n_0w, n_1w, fade_xyzw.z, n_zw);
  Vector2 n_yzw = new Vector2.zero();
  Vector2.mix(n_zw.xy, n_zw.zw, fade_xyzw.y, n_yzw);
  double n_xyzw = mix(n_yzw.x, n_yzw.y, fade_xyzw.x);
  return 2.2 * n_xyzw;
}

// Classic Perlin noise, periodic version
double ashima_pnoise4(Vector4 P, Vector4 rep) {
  Vector4 Pi0 = mod(floor(P), rep); // Integer part modulo rep
  Vector4 Pi1 = mod(Pi0 + ones, rep); // Integer part + 1 mod rep
  Pi0 = mod289(Pi0);
  Pi1 = mod289(Pi1);
  Vector4 Pf0 = fract(P); // Fractional part for interpolation
  Vector4 Pf1 = Pf0 - ones; // Fractional part - 1.0
  Vector4 ix = new Vector4(Pi0.x, Pi1.x, Pi0.x, Pi1.x);
  Vector4 iy = new Vector4(Pi0.y, Pi0.y, Pi1.y, Pi1.y);
  Vector4 iz0 = Pi0.zzzz;
  Vector4 iz1 = Pi1.zzzz;
  Vector4 iw0 = Pi0.wwww;
  Vector4 iw1 = Pi1.wwww;

  Vector4 ixy = permute(permute(ix) + iy);
  Vector4 ixy0 = permute(ixy + iz0);
  Vector4 ixy1 = permute(ixy + iz1);
  Vector4 ixy00 = permute(ixy0 + iw0);
  Vector4 ixy01 = permute(ixy0 + iw1);
  Vector4 ixy10 = permute(ixy1 + iw0);
  Vector4 ixy11 = permute(ixy1 + iw1);

  Vector4 gx00 = ixy00 * (1.0 / 7.0);
  Vector4 gy00 = floor(gx00) * (1.0 / 7.0);
  Vector4 gz00 = floor(gy00) * (1.0 / 6.0);
  gx00 = fract(gx00) - halves;
  gy00 = fract(gy00) - halves;
  gz00 = fract(gz00) - halves;
  Vector4 gw00 = new Vector4.all(0.75) - abs(gx00) - abs(gy00) - abs(gz00);
  Vector4 sw00 = step(gw00, zeros);
  gx00 -= mult(sw00, (step(zeros, gx00) - halves));
  gy00 -= mult(sw00, (step(zeros, gy00) - halves));

  Vector4 gx01 = ixy01 * (1.0 / 7.0);
  Vector4 gy01 = floor(gx01) * (1.0 / 7.0);
  Vector4 gz01 = floor(gy01) * (1.0 / 6.0);
  gx01 = fract(gx01) - halves;
  gy01 = fract(gy01) - halves;
  gz01 = fract(gz01) - halves;
  Vector4 gw01 = new Vector4.all(0.75) - abs(gx01) - abs(gy01) - abs(gz01);
  Vector4 sw01 = step(gw01, zeros);
  gx01 -= mult(sw01, (step(zeros, gx01) - halves));
  gy01 -= mult(sw01, (step(zeros, gy01) - halves));

  Vector4 gx10 = ixy10 * (1.0 / 7.0);
  Vector4 gy10 = floor(gx10) * (1.0 / 7.0);
  Vector4 gz10 = floor(gy10) * (1.0 / 6.0);
  gx10 = fract(gx10) - halves;
  gy10 = fract(gy10) - halves;
  gz10 = fract(gz10) - halves;
  Vector4 gw10 = new Vector4.all(0.75) - abs(gx10) - abs(gy10) - abs(gz10);
  Vector4 sw10 = step(gw10, zeros);
  gx10 -= mult(sw10, (step(zeros, gx10) - halves));
  gy10 -= mult(sw10, (step(zeros, gy10) - halves));

  Vector4 gx11 = ixy11 * (1.0 / 7.0);
  Vector4 gy11 = floor(gx11) * (1.0 / 7.0);
  Vector4 gz11 = floor(gy11) * (1.0 / 6.0);
  gx11 = fract(gx11) - halves;
  gy11 = fract(gy11) - halves;
  gz11 = fract(gz11) - halves;
  Vector4 gw11 = new Vector4.all(0.75) - abs(gx11) - abs(gy11) - abs(gz11);
  Vector4 sw11 = step(gw11, zeros);
  gx11 -= mult(sw11, (step(zeros, gx11) - halves));
  gy11 -= mult(sw11, (step(zeros, gy11) - halves));

  Vector4 g0000 = new Vector4(gx00.x, gy00.x, gz00.x, gw00.x);
  Vector4 g1000 = new Vector4(gx00.y, gy00.y, gz00.y, gw00.y);
  Vector4 g0100 = new Vector4(gx00.z, gy00.z, gz00.z, gw00.z);
  Vector4 g1100 = new Vector4(gx00.w, gy00.w, gz00.w, gw00.w);
  Vector4 g0010 = new Vector4(gx10.x, gy10.x, gz10.x, gw10.x);
  Vector4 g1010 = new Vector4(gx10.y, gy10.y, gz10.y, gw10.y);
  Vector4 g0110 = new Vector4(gx10.z, gy10.z, gz10.z, gw10.z);
  Vector4 g1110 = new Vector4(gx10.w, gy10.w, gz10.w, gw10.w);
  Vector4 g0001 = new Vector4(gx01.x, gy01.x, gz01.x, gw01.x);
  Vector4 g1001 = new Vector4(gx01.y, gy01.y, gz01.y, gw01.y);
  Vector4 g0101 = new Vector4(gx01.z, gy01.z, gz01.z, gw01.z);
  Vector4 g1101 = new Vector4(gx01.w, gy01.w, gz01.w, gw01.w);
  Vector4 g0011 = new Vector4(gx11.x, gy11.x, gz11.x, gw11.x);
  Vector4 g1011 = new Vector4(gx11.y, gy11.y, gz11.y, gw11.y);
  Vector4 g0111 = new Vector4(gx11.z, gy11.z, gz11.z, gw11.z);
  Vector4 g1111 = new Vector4(gx11.w, gy11.w, gz11.w, gw11.w);

  Vector4 norm00 = taylorInvSqrt(new Vector4(dot(g0000, g0000),
      dot(g0100, g0100), dot(g1000, g1000), dot(g1100, g1100)));
  g0000 *= norm00.x;
  g0100 *= norm00.y;
  g1000 *= norm00.z;
  g1100 *= norm00.w;

  Vector4 norm01 = taylorInvSqrt(new Vector4(dot(g0001, g0001),
      dot(g0101, g0101), dot(g1001, g1001), dot(g1101, g1101)));
  g0001 *= norm01.x;
  g0101 *= norm01.y;
  g1001 *= norm01.z;
  g1101 *= norm01.w;

  Vector4 norm10 = taylorInvSqrt(new Vector4(dot(g0010, g0010),
      dot(g0110, g0110), dot(g1010, g1010), dot(g1110, g1110)));
  g0010 *= norm10.x;
  g0110 *= norm10.y;
  g1010 *= norm10.z;
  g1110 *= norm10.w;

  Vector4 norm11 = taylorInvSqrt(new Vector4(dot(g0011, g0011),
      dot(g0111, g0111), dot(g1011, g1011), dot(g1111, g1111)));
  g0011 *= norm11.x;
  g0111 *= norm11.y;
  g1011 *= norm11.z;
  g1111 *= norm11.w;

  double n0000 = dot(g0000, Pf0);
  double n1000 = dot(g1000, new Vector4(Pf1.x, Pf0.y, Pf0.z, Pf0.w));
  double n0100 = dot(g0100, new Vector4(Pf0.x, Pf1.y, Pf0.z, Pf0.w));
  double n1100 = dot(g1100, new Vector4(Pf1.x, Pf1.y, Pf0.z, Pf0.w));
  double n0010 = dot(g0010, new Vector4(Pf0.x, Pf0.y, Pf1.z, Pf0.w));
  double n1010 = dot(g1010, new Vector4(Pf1.x, Pf0.y, Pf1.z, Pf0.w));
  double n0110 = dot(g0110, new Vector4(Pf0.x, Pf1.y, Pf1.z, Pf0.w));
  double n1110 = dot(g1110, new Vector4(Pf1.x, Pf1.y, Pf1.z, Pf0.w));
  double n0001 = dot(g0001, new Vector4(Pf0.x, Pf0.y, Pf0.z, Pf1.w));
  double n1001 = dot(g1001, new Vector4(Pf1.x, Pf0.y, Pf0.z, Pf1.w));
  double n0101 = dot(g0101, new Vector4(Pf0.x, Pf1.y, Pf0.z, Pf1.w));
  double n1101 = dot(g1101, new Vector4(Pf1.x, Pf1.y, Pf0.z, Pf1.w));
  double n0011 = dot(g0011, new Vector4(Pf0.x, Pf0.y, Pf1.z, Pf1.w));
  double n1011 = dot(g1011, new Vector4(Pf1.x, Pf0.y, Pf1.z, Pf1.w));
  double n0111 = dot(g0111, new Vector4(Pf0.x, Pf1.y, Pf1.z, Pf1.w));
  double n1111 = dot(g1111, Pf1);

  Vector4 fade_xyzw = fade(Pf0);
  Vector4 n_0w = new Vector4.zero();
  Vector4.mix(new Vector4(n0000, n1000, n0100, n1100),
      new Vector4(n0001, n1001, n0101, n1101), fade_xyzw.w, n_0w);
  Vector4 n_1w = new Vector4.zero();
  Vector4.mix(new Vector4(n0010, n1010, n0110, n1110),
      new Vector4(n0011, n1011, n0111, n1111), fade_xyzw.w, n_1w);
  Vector4 n_zw = new Vector4.zero();
  Vector4.mix(n_0w, n_1w, fade_xyzw.z, n_zw);
  Vector2 n_yzw = new Vector2.zero();
  Vector2.mix(n_zw.xy, n_zw.zw, fade_xyzw.y, n_yzw);
  double n_xyzw = mix(n_yzw.x, n_yzw.y, fade_xyzw.x);
  return 2.2 * n_xyzw;
}
