import 'package:vector_math/vector_math.dart';
import 'package:vector_math/vector_math_lists.dart';
import 'dart:async';
import 'dart:collection';
import '../voxel/cube_march_quad.dart';

class RoundRobinScheduler {
  static final ListQueue<Completer<int>> _sleepers = new ListQueue<Completer<int>>();
  static int _iterations = 0;
  static Future<int> sleep() async {
    var waiter = new Completer<int>();
    _sleepers.add(waiter);
    var next = _sleepers.removeLast();
    next.complete(_iterations);
    _iterations++;
    return waiter.future;
  }
}

class IteratorYield {
  final int _cycleCount;
  int _currentCount = -1;
  IteratorYield(this._cycleCount);
  FutureOr<int> iterSleep() async {
    _currentCount++;
    if((_currentCount % _cycleCount) == 0)
      await RoundRobinScheduler.sleep();
    return _cycleCount;
  }
}

class Vector3Heap {
  final ListQueue<Vector3> _cache = new ListQueue<Vector3>();
  int _count = 0;
  void Free(Vector3 v) {
    this._cache.add(v);
    _count--;
  }
  Vector3 Alloc() {
    _count++;
    if(_cache.length == 0) {
      return new Vector3.zero();
    }
    else {
      var v = this._cache.removeLast();
      v.setValues(0.0, 0.0, 0.0);
      return v;
    }
  }
  int get Count {
    return _count;
  }
  void Assert(int count) {
    if(Count != count)
      throw new Exception("vector3Heap leak!");
  }
}

final Vector3Heap vector3Heap = new Vector3Heap();

Stream <double> range(int start, int end) async* {
  for(int i = start; i <= end; i++) {
    yield i.toDouble();
  }
}

Stream<Vector3> rangeXYZ(int xStart, int xEnd, int yStart, int yEnd, int zStart, int zEnd) async* {
  int count = vector3Heap.Count;
  try {
    var tmpVector3 = vector3Heap.Alloc();
    await for (var x in range(xStart, xEnd)) {
      await for (var y in range(yStart, yEnd)) {
        await for (var z in range(zStart, zEnd)) {
          tmpVector3.x = x;
          tmpVector3.y = y;
          tmpVector3.z = z;
          yield tmpVector3;
        }
      }
    }
    vector3Heap.Free(tmpVector3);
  } finally {
    vector3Heap.Assert(count);
  }
}

void quadMix(Vector3 v00, Vector3 v10, Vector3 v11, Vector3 v01,
    Vector3 vMix, Vector3 vOut) {
  int count = vector3Heap.Count;
  try {
    final Vector3 v00_10 = vector3Heap.Alloc();
    final Vector3 v01_11 = vector3Heap.Alloc();

    Vector3.mix(v00, v10, vMix.x, v00_10);
    Vector3.mix(v01, v11, vMix.x, v01_11);
    Vector3.mix(v00_10, v01_11, vMix.y, vOut);

    vector3Heap.Free(v00_10);
    vector3Heap.Free(v01_11);
  } finally {
    vector3Heap.Assert(count);
  }
}

Stream<Vector3> subdivideGrid(
    Vector3 v00, Vector3 v10, Vector3 v11, Vector3 v01, int slices) async* {
  int count = vector3Heap.Count;
  try {
    double scale = 1.0 / slices;
    final Vector3 vWalk = vector3Heap.Alloc();
    await for (Vector3 vMix in rangeXYZ(0, slices, 0, slices, 0, 0)) {
      vMix.scale(scale);
      quadMix(v00, v10, v11, v01, vMix, vWalk);
      yield vWalk;
    }
    vector3Heap.Free(vWalk);
  } finally {
    vector3Heap.Assert(count);
  }
}

Stream<Vector3> subdivideQuad(
    Vector3 v00, Vector3 v10, Vector3 v11, Vector3 v01, int slices) async* {
  int count = vector3Heap.Count;
  try {
    double scale = 1.0 / slices;
    final Vector3 vWalk = vector3Heap.Alloc();
    final Vector3 vOffset = vector3Heap.Alloc();
    await for (Vector3 vMix in rangeXYZ(0, slices - 1, 0, slices - 1, 0, 0)) {
      vMix.copyInto(vOffset);
      vOffset.scale(scale);
      quadMix(v00, v10, v11, v01, vOffset, vWalk);
      yield vWalk;
      vOffset.x += scale;
      quadMix(v00, v10, v11, v01, vOffset, vWalk);
      yield vWalk;
      vOffset.y += scale;
      quadMix(v00, v10, v11, v01, vOffset, vWalk);
      yield vWalk;
      vOffset.y -= scale;
      quadMix(v00, v10, v11, v01, vOffset, vWalk);
      yield vWalk;
    }
    vector3Heap.Free(vWalk);
    vector3Heap.Free(vOffset);
  } finally {
    vector3Heap.Assert(count);
  }
}

Stream<Vector3> unitCubeQuadFaces() async* {
  int count = vector3Heap.Count;
  try {
    var v00 = vector3Heap.Alloc();
    var v10 = vector3Heap.Alloc();
    var v11 = vector3Heap.Alloc();
    var v01 = vector3Heap.Alloc();
    var vIter = vector3Heap.Alloc();
    var vOffset = vector3Heap.Alloc();

    for (double scale = -1.0, offset = 1.0; scale <= 1.0;
    scale += 2.0, offset -= 1.0) {
      vOffset.setValues(offset, offset, offset);
      v00.setValues(0.0, 0.0, 0.0);
      v10.setValues(1.0, 0.0, 0.0);
      v11.setValues(1.0, 1.0, 0.0);
      v01.setValues(0.0, 1.0, 0.0);

      v00.scale(scale);
      v10.scale(scale);
      v11.scale(scale);
      v01.scale(scale);

      for (int rotations = 0; rotations < 3; rotations++) {
        v00.copyInto(vIter);
        vIter.add(vOffset);
        yield vIter;
        v10.copyInto(vIter);
        vIter.add(vOffset);
        yield vIter;
        v11.copyInto(vIter);
        vIter.add(vOffset);
        yield vIter;
        v01.copyInto(vIter);
        vIter.add(vOffset);
        yield vIter;
        rotate(v00);
        rotate(v10);
        rotate(v11);
        rotate(v01);
      }
    }

    vector3Heap.Free(v00);
    vector3Heap.Free(v10);
    vector3Heap.Free(v11);
    vector3Heap.Free(v01);
    vector3Heap.Free(vIter);
    vector3Heap.Free(vOffset);
  } finally {
    vector3Heap.Assert(count);
  }
}

class Polygon {
  final int Count;
  Vector3List _vertices;
  Vector3List get vertices { return _vertices; }
  Vector3List _normals;
  Vector3List get normals { return _normals; }

  Polygon(this.Count) {
    this._vertices = new Vector3List(this.Count);
    this._normals = new Vector3List(this.Count);
  }

  int quadMod(int index) {
    return index % Count;
  }

  final Vector3 _vertice = new Vector3.zero();
  Vector3 GetVertice(int index) {
    vertices.store(quadMod(index), _vertice);
    return _vertice;
  }

  final Vector3 _normal = new Vector3.zero();
  Vector3 GetNormal(int index) {
    normals.store(quadMod(index), _normal);
    return _normal;
  }

  final Vector3 t0 = new Vector3.zero();
  final Vector3 t1 = new Vector3.zero();

  UpdateNormals() {
    for(int i = 0; i < 4; i++) {
      int i0 = i;
      int i1 = quadMod(i + 1);
      int i2 = quadMod(i + 2);
      GetVertice(i1).copyInto(t0);
      t0.sub(GetVertice(i0));
      GetVertice(i2).copyInto(t1);
      t1.sub(GetVertice(i0));
      t0.crossInto(t1, GetNormal(i1));
    }
  }

  Stream<Vector3> VerticesStream() async* {
    for(int i = 0; i < Count; i++)
      yield this.GetVertice(i);
  }

  Stream<Vector3> NormalsStream() async* {
    for(int i = 0; i < Count; i++)
      yield this.GetVertice(i);
  }
}

Stream<Quad> Vector3ToQuad(Stream<Vector3> vertices) async* {
  var quad = new Quad();
  int count = 0;
  await for(var v in vertices) {
    var index = count % quad.Count;
    v.copyInto(quad.vertices[index]);
    if((quad.Count - 1) == index)
      yield quad;
    count++;
  }
}

Stream<Triangle> Vector3ToTriangle(Stream<Vector3> vertices) async* {
  var triangle = new Triangle();
  int count = 0;
  await for(var v in vertices) {
    var index = count % triangle.Count;
    v.copyInto(triangle.vertices[index]);
    if((triangle.Count - 1) == index)
      yield triangle;
    count++;
  }
}

class Triangle extends Polygon {
  Triangle(): super(3);
}

class Quad extends Polygon {
  Quad(): super(4);
}

Stream<Vector3> unitCubeQuadFacesSubdivided(int subDivisions) async* {
  var streamQuadVertices = await unitCubeQuadFaces();
  var streamQuads = await Vector3ToQuad(streamQuadVertices);
  await for(var quad in streamQuads) {
    var streamSubdivide = await subdivideQuad(
        quad.vertices[0], quad.vertices[1], quad.vertices[2], quad.vertices[3], subDivisions);
    await for(var v in streamSubdivide)
      yield v;
  }
}

typedef bool IsHot(Vector3 v);
