library vector_math.test.vector3_test;

import 'package:test/test.dart';
import 'dart:async';
import 'package:rad_textures/rad_textures.dart';
import 'package:vector_math/vector_math.dart';
import 'package:rad_textures/rad_textures.dart' as rad;

void testVector3Quantize() {
  var v = new Vector3(1.0, 2.0, 3.0);
  var q = new quantize();
  int encoded = q.encodeVector3BigInt(v);
  var v2 = new Vector3.zero();
  q.decodeVector3BigInt(encoded, v2);
  expect(v2.x, equals(1.0));
  expect(v2.y, equals(2.0));
  expect(v2.z, equals(3.0));
}

FutureOr<int> roundRobinLoop(String label) async {
  int iter = 0;
  for(int i = 0; i < 3; i++) {
    iter = await RoundRobinScheduler.sleep();
    print(label + ": " + iter.toString());
  }
  return iter;
}

FutureOr<int> testRoundRobinScheduler() async {
  var loop1 = roundRobinLoop("label1: ");
  var loop2 = roundRobinLoop("label2: ");
  var list = new List<Future<int>>();
  list.add(loop1);
  list.add(loop2);
  Future.wait(list);
  expect(await loop1, equals(4));
  expect(await loop2, equals(5));
  return 0;
}

FutureOr<int> testRangeXYZ() async {
  int count = 0;
  Vector3 v;
  await for(v in rangeXYZ(0, 2, 0, 2, 0, 2)) {
    count++;
    print(v.toString());
  }
  expect(v.x, equals(2.0));
  expect(v.y, equals(2.0));
  expect(v.z, equals(2.0));
  expect(count, equals(27));

  return 0;
}

FutureOr<int> testSubdivideGrid() async {
  final z = 0.0;
  final w = 1.0;
  //final h = w / 2.0;
  int count = 0;
  var stream = subdivideGrid(
      new Vector3(z, z, z), new Vector3(w, z, z),
      new Vector3(w, w, z), new Vector3(z, w, z),
      2);
  Vector3 v;
  await for(v in stream) {
    count++;
    print(v.toString());
    if(count == 6) {
      expect(v.x, equals(0.5));
      expect(v.y, equals(1.0));
      expect(v.z, equals(0.0));
    }
  }
  expect(count, equals(9));
  expect(v.x, equals(1.0));
  expect(v.y, equals(1.0));
  expect(v.z, equals(0.0));
  return  0;
}

FutureOr<int> testSubdivideQuad() async {
  final z = 0.0;
  final w = 1.0;
  //final h = w / 2.0;
  int count = 0;
  var stream = subdivideQuad(
      new Vector3(z, z, z), new Vector3(w, z, z),
      new Vector3(w, w, z), new Vector3(z, w, z),
      2);
  Vector3 v;
  await for(v in stream) {
    count++;
    print(v.toString());
    if(count == 8) {
      expect(v.x, equals(0.5));
      expect(v.y, equals(0.5));
      expect(v.z, equals(0.0));
    }
  }
  expect(count, equals(16));
  return  0;
}

FutureOr<int> testCubeMarchQuadPlane() async {
  rad.cube_march_quad _cube_march_quad;
  const double _surface_temperature = 0.5;
  final _voxel_array = new rad.voxel_array(new Vector3.zero(), 2);

  _cube_march_quad = new rad.cube_march_quad(_surface_temperature);
  _voxel_array.iterator_voxels(true).forEach((voxel_walker) {
    double val = voxel_walker.z > 0.5 ? 0.0 : 1.0;
    voxel_walker.set(val);
  });

  _voxel_array.iterator_voxels(false).forEach((voxel_walker) {
    var quads = _cube_march_quad.get_surface_quads(voxel_walker);
    print("quad_vertex_count: ${quads.length}");
    expect(quads.length, equals(4));
    quads.forEach((v) {
      expect(v.z, equals(1.0));
    });
  });

  return 0;
}

FutureOr<int> testCubeMarchCornerPlanes() async {
  rad.cube_march_quad _cube_march_quad;
  const double _surface_temperature = 0.5;
  final _voxel_array = new rad.voxel_array(new Vector3.zero(), 2);
  _cube_march_quad = new rad.cube_march_quad(_surface_temperature);
  _voxel_array.set(0, 0, 0, 1.0);

  _voxel_array.iterator_voxels(false).forEach((voxel_walker) {
    var quads = _cube_march_quad.get_surface_quads(voxel_walker);
    print("quad_vertex_count: ${quads.length}");
    expect(quads.length, equals(12));
    quads.forEach((v) {
      bool is_origin = (v.x == 0.0) && (v.y == 0.0) && (v.z == 0.0);
      expect(is_origin, equals(false));
    });
  });

  return 0;
}

FutureOr<int> testMarchUnitCubeFacesCalcCenter() async {
  Vector3 v;
  Vector3 sum = new Vector3.zero();
  int count = 0;
  await for(v in unitCubeQuadFaces()) {
    sum.add(v);
    count++;
  }
  double scale = 1.0 / count;
  sum.scale(scale);
  const double mid = 0.5;
  expect(sum.x, equals(mid));
  expect(sum.y, equals(mid));
  expect(sum.z, equals(mid));
  print("unit cube faces vert count: " + count.toString());
  expect(count, equals(24));

  return 0;
}

FutureOr<int> testMarchUnitCubeFaceCenters() async {
  var uniqueFaceCenters = new Set<String>();
  Vector3 v;
  Vector3 sum = new Vector3.zero();
  Vector3 center = new Vector3(0.5, 0.5, 0.5);
  const double mid = 0.5;
  double scale = 1.0 / 4.0;
  int count = 0;
  print("march unit cube face centers:");
  await for(v in unitCubeQuadFaces()) {
    print(v.toString());
    sum.add(v);
    count++;
    if((count > 0) && (count % 4 == 0)) {
      sum.scale(scale);
      print("face center: " + sum.toString());
      sum.sub(center);
      expect(sum.length, equals(mid));
      uniqueFaceCenters.add(sum.toString());
      sum.setValues(0.0, 0.0, 0.0);
    }
  }
  expect(uniqueFaceCenters.length, equals(6));
  print("unit cube face centers vert count: " + count.toString());
  expect(count, equals(24));

  return 0;
}

FutureOr<int> testMarchUnitCubeFaceCentersSubdivided() async {
  var streamVertices = await unitCubeQuadFacesSubdivided(2);
  await for(var vertice in streamVertices)
    print("vertice: " + vertice.toString());
  return 0;
}

void main() {
  group('Quantize', () {
    test('Vector3Quantize', testVector3Quantize);
  });
  group('RoundRobin', () {
    test('Scheduler', testRoundRobinScheduler);
  });
  group('Mesh', () {
    test('RangeXYZ', testRangeXYZ);
    test('SubdivideGrid', testSubdivideGrid);
    test('SubdivideQuad', testSubdivideQuad);
  });
  group('MarchCube', () {
    test('CubeMarchQuadPlane', testCubeMarchQuadPlane);
    test('CubeMarchCornerPlanes', testCubeMarchCornerPlanes);
  });
  group('UnitCubeQuadFaces', () {
    test('MarchUnitCubeFacesCalcCenter', testMarchUnitCubeFacesCalcCenter);
    test('MarchUnitCubeFaceCenters', testMarchUnitCubeFaceCenters);
    //todo, test('MarchUnitCubeFacesSubdivided', testMarchUnitCubeFaceCentersSubdivided);
  });
}
